;*************************************************
; Module: mouse.a51
;
; Firmware design and code for I2C desktop bus Mouse
;   Environment:  83C751  Assembler
;   Author:       Robert Clemens   10-Jul-1990
;                 (I2C I/O adapted from P.Sichel's "Keyboard" code)  
;   Revision:     6-Mar-1991
;
;   31-Jan-1991 PAS  Add numerous keyboard fixes.
;                    Streamline input sample and I2C code.
;                    Separate HW dependent constants.
;                    Fix RxEnable after bus time out.
;
;   06-Feb-1991 PAS  Rev X0.6
;                    Fix sample timer initialization, use 14ms as default.
;                    Fix length checking to allow commands with
;                      more parameters than required.
;                    Implement Set Interval command.
;                    Handle LLLLL=0 to mean 32.
;                    Fix ARL during self-addressed reset message.
;                    Fix to handle DRDY and ARL together.
;                    Re-order mouse buttons as MRL, update Capabilites.
;                    Do not allow other interrupts during TimerI svc.
;                    Document sampling requirements.
;                    Document hardware details.
;                    Add check to skip waiting after DNRXB.
;                    Misc clean-up in: BeMast, Assign,...
;                    Use include files for 751 registers and ODB msgs.
;
;    6-Mar-1991 PAS  Rev X0.7
;                    Fix MN8Bit to check ARL before clearing DRDY.
;                    Separate SendRpt flag from movement detected (Movement)
;                       so only Timer0 will initiate motion reports.
;                    Don't send Position report to def_addr even if polled.
;                    Report InputError for invalid checksum,
;                       unrecognized command code, or illegal parameter
;                       value.  Do not complain about parameters beyond
;                       those anticipated.
;                    Sample quadrature inputs between I2C bytes
;                       to insure accurate tracking.
;
;   13-Mar-1991 PAS  Fix DoStp4 to borrow Rx code and become IDLE rcv.
;
;   26-Mar-1991 PAS  Rex X0.8
;                    Update to use 4-byte device number.
;                    Get new device number only after Reset.
;
;    8-Apr-1991 PAS  Add error checking to CapReq.
;
;    9-Jul-1991 PAS  Add protocol_revision as part of module revision.
;
;   29-Jul-1991 PAS  Protocol_revision in 1-byte, fix ARL and MASTER bug.
;                    Ignore unrecognized commands.  Add I_MsgCheck.
;
;   11-Sep-1991 PAS  Rev X1.2.  Identify button positions as 1,2,3.
;                    Use new ab.inc file.  Change I_MsgCheck to
;                    I_Error and do not overwrite pending SndType.
;                    Retry message after Negative Ack (NACnt).
;
;    7-Oct-1991 PAS  Rev X1.3.  Fix spurious large count problem.
;                    Fix STOP detected while MASTER without ARL.
;
;   25-Oct-1991 PAS  Rev X1.4.  Improve TimerI handler to avoid
;                    lockup when MASTRQ with SCL low.
;
;    4-Nov-1991 PAS  V1.0 release for Boston mfg.
;   22-Dec-1991 PAS  V1.1 align data with tx bit that lost arbitration
;
;
;************************************************
; ACCESS.bus Technology Transfer Agreement
; 
; Grant of Right to Use ACCESS.bus Technology
; 
; In exchange for your agreeing to the warranty disclaimer and liability
; limitation stipulated in this Technology Transfer Agreement, Digital
; Equipment Corporation (Digital) grants at no cost to you a royalty-free
; nonexclusive license to use ACCESS.bus technology (as specified in the
; ACCESS.bus Hardware and Protocol Specifications) to design and develop
; any kind of option board, computer system, or application-specific
; integrated circuit (ASIC).  This Agreement does not grant you any other
; rights in Digital's patents, copyrights, trade secrets, trademarks, or
; licenses to ACCESS.bus technology.
; 
; Warranty Disclaimer
; 
; The ACCESS.bus technology is transferred "as is."  Digital
; expressly disclaims all implied warranties including, but not
; limited to, the implied warranties of merchantability and fitness
; for a particular purpose.  Digital does not warrant, guarantee, or
; make any representations regarding the use of or the results of the
; use of source code and related documents in terms of correctness,
; accuracy, or reliability.  Digital believes the code is accurate;
; however, you must assume the risk as to the results and performance
; of any product you develop that is based on the ACCESS.bus technology.
; 
; Limits of Liability
; 
; You agree that Digital shall not be liable to you under this Agreement for
; any damages, including without limitation any lost profits or lost savings,
; or any consequential, incidental, or punitive damages arising out of the use
; or inability to use the ACCESS.bus Mouse Device source code and related
; documents, or for any claim by another party.  Your exclusive remedy under
; this Agreement shall be the furnishing by Digital of the technical support
; provided herein.  You agree to hold Digital harmless for all claims and
; damages arising from any third party as a result of their use of or inability
; to use any product you develop based on ACCESS.bus technology.
;************************************************
$TITLE (Digital ACCESS.bus Mouse, V1.1)
$DATE (12/22/91)
$DEBUG


; Symbolic addresses and masks
 
$INCLUDE(reg751.inc)
$INCLUDE(ab.inc)
 
$EJ
;***************************************
;* Hardware/timing-dependent constants *
;***************************************

;CT             EQU     02h     ;CT1, CT0 fmax = 16.8 MHz
;CT             EQU     01h     ;CT1, CT0 fmax = 14.25 MHz
;CT             EQU     00h     ;CT1, CT0 fmax = 11.7 MHz
CT              EQU     03h     ;CT1, CT0 fmax = 9.14 MHz
 
IntEnab         EQU     09Ah    ;enable EA+EI2+ET1+ET0.
INIT_TCON       EQU     010h    ;Timer0 init for internal operation.

DEF_RTH         EQU     0DBh    ;Sampling interval (14ms with 8 MHz clock).
DEF_RTL         EQU     08Bh    ; Used as default Timer0 reload value.
MSECH           EQU     002h    ;Timer offset for 1 ms with 8 MHz clock.
MSECL           EQU     09Ah    ; 029Ah=666 * 3/2 mms/tick = 1000 ticks = 1ms.

DelayATN        EQU     8       ;about 50uS (wait for I2C activity)

CapFragLen      EQU     16      ;Capabilities fragment length.

; *** Hardware Interface Notes ***
; P3 is used to read the X-Y quadrature inputs.
;   P3.0 = XB 
;   P3.1 = XA
;   P3.2 = YA
;   P3.3 = YB
;
; Notice P3.0 is connected to the B side of the X encoder
;  while P3.2 is connected to the A side of the Y encoder.
; This is to compensate for the orientation of the inclined cyclinders.
; Positive X movement causes clockwise rotation of the encoder shaft.
; Positive Y movement causes counter clockwise rotation of the encoder shaft.
;
; P1 is used to read the switch inputs.
;   P1.0 = middle mouse button.
;   P1.1 = left mouse button.
;   P1.2 = right mouse button.
;
; 6-Feb-1991  This order of buttons reflects the current PCB layout
;             and is subject to change.

$EJ
; Locator messages (device defined)
;   send types
LD_Position     EQU     3       ;Device state report

;   receive types
LD_Poll         EQU     0B0h
LD_SetInterval  EQU     082h


; SelfTest errors (0=success)
NO_ERROR        EQU     0
ROM_ERROR       EQU     1
RAM_ERROR       EQU     2


$EJ
;************
; RAM usage *
;************
; I2C variables
; Register bank 0
                           ;R0 - command parameter
                           ;R1 - index pointer
I2CDat   DATA   02h        ;The byte being sent or received.
BitCnt   DATA   03h        ;I2C bit counter
ByteCnt  DATA   04h        ;I2C message byte counter
ATNCnt   DATA   05h        ;ATN Retry counter
I2CCxt   DATA   06h        ;I2C context, the event the CPU is waiting for.
Temp     DATA   07h        ;All purpose temp

; Desktop Bus Protocol
MyAddr   DATA   08h        ;I2C address assigned this device.
NACnt    DATA   09h        ;Negative Ack retry counter.
RcvType  DATA   0Ah        ;Message or command type being received.
SndType  DATA   0Bh        ;Message type being sent, or pending.
MsgLen   DATA   0Ch        ;Message length field.
Check    DATA   0Dh        ;Message checksum.
RandH    DATA   0Eh        ;Random number (2 bytes)
RandL    DATA   0Fh

; Locator report buffer
ReportBuf       EQU     10h     ;Beginning of position report buffer.
Switch2         DATA    10h     ;Switch data (Buttons 9 to 16)
Switch1         DATA    11h     ;Switch data (Buttons 1 to 8 )
XBUF2           DATA    12h     ;XData transmission buffer (MSB)
XBUF1           DATA    13h     ;XData transmission buffer (LSB)
YBUF2           DATA    14h     ;YData transmission buffer (MSB)
YBUF1           DATA    15h     ;YData transmission buffer (LSB)

XCOUNT          DATA    16h     ;XData
YCOUNT          DATA    17h     ;YData
MARKER          EQU     YCOUNT+1

CapOffset       DATA    18h     ;Capabilities fragment offset
CapLen          DATA    19h     ;Capabilities fragment length
SelfTest        DATA    1Ah     ;Self Test result variable location
RomSum          DATA    1Bh     ;Hold ROM checksum
SampleClock     DATA    1Ch     ;Time stamp of last sample
                                ;3 spare.

; Bit addressable area begins at 20h
; Location and switch compare variables
LastXY          DATA    20h     ;Last X/Y Position  (from P3)
LastSW          DATA    21h     ;Last Switch Status (from P1)
TranXY          DATA    22h     ;Port 3 transition register (bit addressable)

$EJ
; I2C status and position scanning flags.
Flags           DATA    23h
Prot            BIT   Flags.0   ;1=C/S message; 0=device data stream.
SendRpt         BIT   Flags.1   ;New Position Report flag.
Movement        BIT   Flags.2   ;Movement detected flag.
RxEnable        BIT   Flags.3   ;I2C receive enable.
TxSelfRst       BIT   Flags.4   ;Indicates send Self-Reset after Assign (0).
KeepID          BIT   Flags.5   ;Set means keep same device number.
NotMyID         BIT   Flags.6
AA              BIT   Flags.7   ;Assert Acknowledge.

FlagsA          DATA    24h
MsgCheck        BIT   FlagsA.0  ;I2C message checksum or framing error.

                                ;11 spare

StackBase       DATA    30h     ;16 byte stack
                                ; need 2 bytes per subroutine
                                ; 4 bytes per interrupt (max 2)
;RAM ends at 3Fh
; 50 bytes RAM used, 64 maximum.


$EJ
;******************
; Code begins
;******************
        ORG     0       ;Power up reset vector.
        AJMP    PwrUp

        ORG     003h    ;INT0 is not used.
        AJMP    PwrUp
;
; Use this spare byte to hold the ROM checksum complement.
        DB      000h

        ORG     00Bh    ;Timer0 interrupt vector
        AJMP    Timer0  ;16 bit system tick generator

        ORG     013h    ;INT1 is not used.
        AJMP    PwrUp

        ORG     01Bh    ;TimerI interrupt vector
        AJMP    TimerI  ;I2C time out timer

        ORG     023h    ;I2C interrupt vector
        AJMP    I2CINT


$EJ
;*****************************
; Timer0 Interrupt
; Position sampling interval time out.
;*****************************

Timer0: PUSH    PSW                     ;Save registers we need.
        PUSH    ACC
        MOV     A,MyAddr                ;Check for default address.
        CJNE    A,#Adr_Default,AddrOK
        SJMP    T0Exit                  ;Don't send Position reports
                                        ; to default address.
AddrOK: MOV     A,P1                    ;Get switch info from P1
                                        ; (0=button depressed).
        CPL     A                       ;Complement
        ANL     A,#00000111b            ;We need low 3 bits of P1
        CJNE    A,LastSW,CHNSWI         ;if new, switches did change
        ;Switches did not change, check movement.
        JNB     Movement,T0Exit
        SJMP    T0Send                  ;Yes, go send report.

CHNSWI: MOV     LastSW,A                ;Save LastSW for next compare.
        ;Re-order switches from RLM to MRL until PCB is fixed.
        RRC     A
        MOV     ACC.2,C
        ANL     A,#00000111b
        MOV     Switch1,A               ;Move to output buffer.

T0Send: SETB    SendRpt                 ;Set to send Position report.
        SETB    I2CFG.MASTRQB           ;Request to be master.

T0Exit: POP     ACC
        POP     PSW
        RETI

$EJ
;*******************************
; TimerI interrupt 
; The I2C bus has timed out,
; no SCL for at least 1020 machine cycles during an active frame.
; Since SCL is stuck, we can't wait for DRDY.
; Try to fix it manually.
;*******************************

TimerI:   CLR   IE.EA           ;Disable interrupts.
          MOV   I2CFG,#CLRTI+CT ;Clear interrupt and turn off TimerI.
                                ; manually clear SLAVEN & MASTRQ.
          MOV   I2CON,#CXA+CARL+CDR+CSTR+CSTP   ;Clear I2C flags.
          MOV   SP,#StackBase   ;Reset SP.

; Attempt to regain control of the I2C bus after a bus fault.
FixBus:   SETB  SCL             ;Insure I/O port is not locking I2C.
          SETB  SDA
          JNB   SCL,ResetBus    ;If SCL is low, bus cannot be fixed.
          JB    SDA,RStop       ;If SCL & SDA are high, force a stop.
          ;SDA is low, attempt to release SDA by clocking SCL.
          MOV   BitCnt,#9       ;Set max # of tries to clear bus.
ClockBus: CLR   SCL             ;Force an I2C clock.
          ACALL SDelay
          JB    SDA,RStop       ;Did it work?
          SETB  SCL
          ACALL SDelay
          DJNZ  BitCnt,ClockBus ;Repeat clocks until SDA clears or retry limit.
          SJMP  ResetBus        ;Failed to fix bus by this method.

RStop:    CLR   SDA             ;Try forcing a stop since
          ACALL SDelay          ; SCL & SDA are both high.
          SETB  SCL
          ACALL SDelay
          SETB  SDA
;
ResetBus: ;Wait for bus to clear.
          JNB   SCL,$
          JNB   SDA,$
          MOV   A,#1                    ;Pause for bus to stabilize.
          ACALL LDelay
          ;Re-enable I2C functions.
          MOV   SndType,#I_NoMsg        ;Cancel message if any.
          SETB  RxEnable                ;Enable receiving.
          ACALL InitI2C                 ;Initialize I2C.
          AJMP  MAIN                    ;Restart MAIN.


; Initialize I2C functions
InitI2C: MOV    I2CFG,#SLAVEN+TIRUN+CT  ;Enable I2C
        ;Set I2C to be idle receiver & clear all flags.
        MOV     I2CON,#CXA+IDLE+CDR+CARL+CSTR+CSTP
        MOV     I2CCxt,#RXIDLE          ;Context idle receiver
        ACALL   XRETI                   ;Clear pending interrupt if any.
        MOV     IE,#IntEnab             ;Enable interrupts (EA+EI2+ET1+ET0)
        MOV     A,#16                   ;Wait to sync message frame.
        ACALL   LDelay
        RET

; Long Delay, A/2 milliseconds.
LDelay: MOV     R7,#166
        DJNZ    R7,$
        DJNZ    ACC,LDelay
        RET

; Short delay routine (10 machine cycles).
SDelay:   ACALL SD1
SD1:      NOP
          NOP
          RET

$EJ
;***********************************
; Power up initializion starts here
;***********************************
; Reset command branches to here.
PwrUp:  MOV     MyAddr,#Adr_Default     ;Re-initialize default address.
        CLR     A
        MOV     IE,A            ;Disable interrupts.
        MOV     I2CFG,#CT       ;Disable I2C
        ANL     PSW,#0E7h       ;Select RB0.
        MOV     SP,#StackBase
        ; Initialize I/O pins
        MOV     P0,#03h         ;Initialize I2C I/O pins, SCL & SDA high
        MOV     P1,#0FFh        ;P1 set for input to read switches.
        MOV     P3,#0FFh        ;P3 set for input to read X-Y.
        ; Initialize Timer0
        MOV     RTH,#DEF_RTH    ;14 ms at 8 MHz
        MOV     RTL,#DEF_RTL
        MOV     TCON,#INIT_TCON ;Running, internal mode clock/12.

;************
; Perform ROM Test (0-7FFh, 2K bytes)
;************
TestROM:MOV     B,#0            ;Initialize sum
        MOV     DPTR,#0000h     ;Set pointer to start of ROM
        CLR     C
SumLp:  CLR     A
        MOVC    A,@A+DPTR       ;Get byte from ROM
        ADDC    A,B             ;Add sum
        MOV     B,A             ;Save sum in B
        INC     DPTR
        MOV     A,DPH           ;Check if ROM complete
        CJNE    A,#08h,SumLp
        JNC     TestSum         ;Add carry if set
        INC     B
TestSum:MOV     A,B
        JZ      TestRAM         ;If zero, ROM is Okay
        MOV     SelfTest,#ROM_ERROR
        MOV     RomSum,A        ;Save bad checksum.
        SJMP    BadMem1

$EJ
;************
; Perform RAM test (0-3Fh, 64 bytes)
; Does not test special function registers.
; A, B, and R0 are not preserved.
;************
TestRAM: MOV    R0,#0AAh        ;Test RAM location 0.
        CJNE    R0,#0AAh,BadMem
        MOV     R0,#055h
        CJNE    R0,#055h,BadMem
        MOV     R0,#3Fh         ;Init R0 to top of RAM.
        MOV     A,#0AAh         ;Test alternate bits.
ChkRAM: MOV     B,@R0           ;Save previous contents.
        MOV     @R0,A
        CJNE    @R0,#0AAh,BadMem
        RL      A               ;Test other bits.
        MOV     @R0,A
        CJNE    @R0,#055h,BadMem
        RL      A
        MOV     @R0,B           ;Restore contents.
        DJNZ    R0,ChkRAM
        SJMP    MemOK
BadMem: MOV     SelfTest,#RAM_ERROR

; Report bad memory.  Since a memory problem was detected, the
; normal I2C transmit code may be unreliable.  Hope it isn't
; a fatal problem and use it anyway.  There's only so much
; we can do.  Could add special code here.
BadMem1: SJMP   InitRAM

MemOK:  MOV     SelfTest,#0
        SJMP    InitRAM


$EJ
;************
;Initialize RAM
;************
InitRAM:CLR     A                       ;Init Acc to Zero
        MOV     RcvType,A               ;Clr RcvType
        MOV     SndType,A               ;Clr SndType
        MOV     R1,#ReportBuf           ;Clear report buffer & compare vars
ClrBuf: MOV     @R1,A                   ;Clear location
        INC     R1                      ;Go to next location
        CJNE    R1,#Marker,ClrBuf       ;Check for end
        MOV     Flags,A                 ;Init Flags
        MOV     LastSW,A                ;Init last switch image.
        MOV     NACnt,#5                ;Negative Ack retry count.

;Init LastXY
        MOV     A,P3                    ;Read X-Y quad inputs to init LastXY.
        ANL     A,#00Fh                 ;Low 4 bits only.
        MOV     LastXY,A                ;Init LastXY

;************
;Set up to transmit self test report
;************
        CLR     RxEnable                ;Disable receiving.
SetUp:  ACALL   InitI2C

Report: MOV     I2CCxt,#RXIDLE          ;Set context idle receiver.     ;*RC*
        MOV     SndType,#I_Attn
        SETB    I2CFG.MASTRQB           ;Request to be master.
        JB      I2CFG.MASTRQB,$         ;Wait for message sent.         ;*PAS*
        MOV     A,SelfTest
        JZ      RepDn                   ;Go if Selftest OK.

        ;Selftest failed.
        ;Send 2nd report and try to start anyway.
        MOV     SelfTest,#NO_ERROR
        SJMP    Report

RepDn:  SETB    RxEnable                ;Enable Receiver and
                                        ; fall through to MAIN.

$EJ
;****************************************************************************
; Main Routine 
;
; Sample X-Y quadrature inputs and compute mouse movement.
; Accuracy requirement is:  +/- 3%  0-10 inches per second.
; 10 inches per second @ 200 dpi means up to 2000 input changes/second.
; Minimum (Nyquist) sampling rate is 4000/sec or sample every 250s.
; This is only two character times at 80k bps.
; For best accuracy, should sample between every I2C character.
;
; Sample timing with 8 MHz crystal:
;   no transition:      6 cycles        9s
;   X or Y transition:  25-31 cycles    38-47s
;   X and Y transition: 42-46 cycles    63-69s
;
;****************************************************************************

MAIN:
        MOV     A,TL            ;Read timer to wait at least
        SUBB    A,SampleClock   ; 64 cycles (96 sec) between samples.
        JNB     ACC.6,MAIN
        ;Take a sample
        ;CLR    IE.EA           ;Protect sample code from interrupts.
        MOV     SampleClock,TL
        ACALL   Sample          ;Sample X-Y quadrature inputs.
        ;SETB   IE.EA
        SJMP    MAIN

; Sample X-Y quadrature inputs and update X-Y counters.
;   b3-b0 = YB YA XB XA.
;
;   Channel A:   0 1 1 0 0 1 1 0 0     --->positive movement
;   Channel B:   0 0 1 1 0 0 1 1 0     <---negative movement

;   A and C are not preserved.
;
Sample: MOV     A,P3            ;Read X & Y position detectors.
        ANL     A,#00001111B    ;We only need the low 4 bits
        CJNE    A,LastXY,TRAN   ;Compare to last image.
        RET                     ; If no change, return.

TRAN:   MOV     TranXY,LastXY   ;Set up to calculate XY transition.
        MOV     LastXY,A        ;Save new P3 image.
        XRL     TranXY,A        ;Mark transition bits (1=changed).

XPULSE: JB      TranXY.0,XA     ;Branch on bit transitions.
        JB      TranXY.1,XB
YPULSE: JB      TranXY.2,YA
        JB      TranXY.3,YB
        RET

        ;Decode direction from quadrature.
        ;Change in XA
XA:     MOV     A,LastXY
        ANL     A,#00000011B    ;Get X' X
        JZ      XDEC            ;If 00, backward
        XRL     A,#00000011B
        JZ      XDEC            ;If 11, backward
        SJMP    XINC            ;01 or 10, forward.
        ;Change in XB
XB:     MOV     A,LastXY
        ANL     A,#00000011B    ;Get X' X
        JZ      XINC            ;If 00, forward
        XRL     A,#00000011B
        JZ      XINC            ;If 11, forward
        ;01 or 10, backward, fall through to decrement

XDEC:   MOV     A,#080h         ;Do not decrement if
        XRL     A,XCOUNT        ; count already at minimum -127.
        JZ      XDEC2
        DEC     XCOUNT
        SETB    Movement        ;Note position has changed.
XDEC2:  SJMP    YPULSE          ;Check for possible Y pulse.

XINC:   MOV     A,#07Fh         ;Do not increment if
        XRL     A,XCOUNT        ; count already at maximum 127.
        JZ      XINC2
        INC     XCOUNT
        SETB    Movement        ;Note position has changed.
XINC2:  SJMP    YPULSE          ;Check for possible Y pulse.

        ;Change in YA
YA:     MOV     A,LastXY
        ANL     A,#00001100B    ;Get Y' Y
        JZ      YDEC            ;If 00, backward
        XRL     A,#00001100B
        JZ      YDEC            ;If 11, backward
        SJMP    YINC            ;01 or 10, forward.
        ;Change in YB
YB:     MOV     A,LastXY
        ANL     A,#00001100B    ;Get Y' Y
        JZ      YINC            ;If 00, forward
        XRL     A,#00001100B
        JZ      YINC            ;If 11, forward
        ;01 or 10, backward, fall through to decrement.

YDEC:   MOV     A,#080h         ;Do not decrement if
        XRL     A,YCOUNT        ; count already at minimum -127.
        JZ      YDEC2
        DEC     YCOUNT
        SETB    Movement        ;Note position has changed.
YDEC2:  RET

YINC:   MOV     A,#07Fh         ;Do not increment if
        XRL     A,YCOUNT        ; count already at maximum 127.
        JZ      YINC2
        INC     YCOUNT
        SETB    Movement        ;Note position has changed.
YINC2:  RET

$EJ
                ; I2C message processing contexts:
RXIDLE  EQU     1                 ;Idle receiver waiting for start.
RXBIT   EQU     2                 ;Waiting to receive a bit.
RXACK   EQU     3                 ;Waiting for ACK to complete.

TXBIT   EQU     4                 ;Waiting to send a bit.
TXREAD  EQU     5                 ;Waiting to read ACK.
TXACK   EQU     6                 ;Waiting for ACK.



; I2C Interrupt Reasons
;   Events CPU might be waiting for
;     Receive
;     (1) Start signal detected by idle slave (DRDY)
;     (2) Next bit received (DRDY)
;     (3) Acknowledge has been sent (DRDY)
;     Transmit
;     (4) Bus mastership granted (START and MASTER)
;     (5) Ready to transmit next bit (DRDY)
;     (6) Acknowledge received (DRDY)
;   Unsolicited
;   (7) Arbitration loss (ARL)
;   (8) Sender aborted message (STOP)
;   (9) Sender started new message before slave became idle (START)
;
; Only some of these events can occur at any time depending on
; the state of sending or receiving a message.
;
; When the interrupt occurs, the keyboard needs to recover
; the context from which it was sending or receiving a
; message.  This context is maintained as follows:
;
;   I2Ccxt   I2C context, what event is expected.
;   I2CDat   The byte being sent or received.
;   BitCnt   Where we are in the sending or receiving the byte.
;   ByteCnt  Where we are in sending or receiving a message
;   Check    Computed checksum.
;   RcvType  The message or command type being received.
;              This may determine how successive bytes
;              are to be processed.
;   SndType  Type of message being sent or pending.
;              This will determine how bytes are transmitted.
;   SendRpt   Flag indicating CPU is waiting to send a Position
;              report (and requested to become master).



$EJ
;*********************
; Enter I2C Interrupt
;*********************
I2CINT: CLR     IE.EI2                  ;disable the I2C interrupt
        ACALL   XRETI                   ;then re-enable others
        PUSH    PSW                     ;save registers
        PUSH    ACC

; Dipatch interrupt
DISPAT: JNB     I2CON.MST,SLAVE
        AJMP    MASTER                  ;go if we're master

; Wait for ATN
WaitATN: MOV    R5,#DelayATN            ;Load ATN count (about 50mms)
Wait1:  JB      I2CON.ATN,DISPAT        ;If ATN, dispatch next event,
        DJNZ    R5,Wait1                ; else loop to try again.
        ;If not seen after count tries,
        ; return from I2C interrupt.

; Exit I2C interrupt
; restore registers and return from interrupt
I2CRTI: POP     ACC
        POP     PSW
        SETB    IE.EI2                  ;re-enable I2C interrupt
        RET                             ;return to interrupted process

XRETI:  RETI                            ;used at start of service routine


$EJ
;*************************
; SLAVE RECEIVER
;   R2=I2CDat, R3=BitCnt, R4=ByteCnt, R5=ATNCnt, R6=I2CCxt
;*************************
; Handle DRDY
;*************************
SLAVE:  JNB     I2CON.DRDY,NDRDY        ;Is it DRDY?
; - context waiting for bit
        CJNE    R6,#RXBIT,NRxBit        ;Context waiting for bit?
Rx0:    MOV     A,R2                    ;Yes, get data in A
Rx1:    DJNZ    R3,N8Bit                ;8th bit?

        ; Read 8th bit
        MOV     C,I2CON.RDAT            ;Get 8th bit, don't clear ATN
        RLC     A                       ;Include the 8th bit
        MOV     R2,A                    ;Put data in R2.
        XRL     Check,A                 ;XOR it to check

        ;Send Acknowledge as appropriate.
        CJNE    R4,#0,DoAck1            ;Address byte? (ByteCnt=0)
        CJNE    A,MyAddr,NotMe          ;Is it my address?
        MOV     Check,A                 ;Yes, initialize check
        CLR     AA                      ;AA=Flags.7
DoAck1: MOV     I2DAT,Flags             ;Assert Acknowledge (AA=Flags.7)
        MOV     R6,#RXACK               ;Set context waiting for ACK
        JB      I2CON.DRDY,AckCmp       ;Can we skip waiting?
        AJMP    WaitATN                 ;Wait for ATN.
; Not addressed to me
NotMe:  MOV     R6,#RXIDLE              ;Set context to be idle receiver
        MOV     I2CON,#CDR+CSTR+CSTP+CARL+IDLE
        AJMP    I2CRTI                  ;Resume interrupted activity

        ; Read bits 1-7
N8Bit:  CLR     ACC.7
        ORL     A,I2DAT                 ;Include the bit, clear ATN.
        RL      A                       ;Data comes in at MSB.
        JB      I2CON.DRDY,Rx1          ;If DRDY, short cut.
        JB      I2CON.DRDY,Rx1
        JB      I2CON.DRDY,Rx1          ;One more try.
        MOV     R2,A                    ;Put data back.
        AJMP    WaitATN

; - context waiting for ACK to complete
NRxBit: CJNE    R6,#RXACK,NRxAck        ;Context waiting for ACK?
AckCmp: MOV     I2CON,#CDR+CXA          ;ACK complete, clr xmt.
        ;Process complete byte.
        AJMP    DORXB
        ; Return is AJMP DNRXB.
DNRXB:  INC     R4                      ;Increment ByteCnt.
        ;ACALL  Sample
SetRXB: MOV     R6,#RXBIT               ;Set context for next byte.
        MOV     R2,#0                   ;Clear receive buffer
        MOV     R3,#8                   ;BitCnt=8
        JB      I2CON.DRDY,Rx0          ;If DRDY, short cut.
        AJMP    WaitATN                 ;Wait for ATN

; - context idle slave
NRxAck: CJNE    R6,#RXIDLE,NRxIdle      ;Context idle slave?
        JNB     RxEnable,NotMe          ;Am I enabled?
        ;Yes, initialize to receive first byte
        MOV     R3,#7                   ;BitCnt=7 (remaining)
        CLR     A                       ;Data in A
        MOV     R4,A                    ;ByteCnt=0
        MOV     R6,#RXBIT               ;I2CCxt=receive next bit
        AJMP    N8Bit
; Context was not waiting for Next bit, ACK, or Idle slave.
; Could be ARL.  Dispatch other flags.
; Do not clear DRDY, we'll come back after ARL if necessary.
NRxIdle: 


$EJ
;*********************************************
; It wasn't DRDY, could be ARL, START, or STOP
; Handle ARL
;*********************************************
NDRDY:  JNB     I2CON.ARL,RxStop        ;Is it ARL?
        ;Yes, note MASTRQ is still on unless we were sending STOP.
        ;SndType is still pending.  If it was a position report,
        ;indicate pending report in SendRpt in case SndType is needed.
ARL0:   JNB     I2CFG.MASTRQB,Naddr     ;Was I sending STOP?
        MOV     A,SndType
        CJNE    A,#LD_Position,ARL01    ;Was I sending position?
        MOV     SndType,#I_NoMsg
        SETB    SendRpt
ARL01:  CJNE    A,#I_Error,ARL1         ;Was I sending Error?
        MOV     SndType,#I_NoMsg
        SETB    MsgCheck
ARL1:   JNB     RxEnable,NAddr          ;Am I enabled?
        JB      I2CON.STR,SlvStart      ;Handle start.
        CJNE    R4,#0,ARL3              ;Did we ARL in Address (ByteCnt=0)?
; Lost arbitration in address, set context to read rest
; of address in case message is to me.
ARL2:   MOV     I2CON,#CARL             ;Clear ARL.
        MOV     R6,#RXBIT               ;Set context waiting to read bit.
        JB      I2CON.DRDY,Rx0          ;If DRDY, short cut.
        AJMP    WaitATN
; Lost arbitration outside dst addr
ARL3:   CJNE    A,#I_Reset,NAddr        ;Was I sending I_Reset to my Addr?
        CLR     TxSelfRst               ;Yes, reset flag since I lost.
        CLR     AA                      ;Acknowledge received bytes.
        SJMP    ARL2                    ;Message must be for me.
; Message not for me, go back to idle receive.
NAddr:
IdleS:  MOV     I2CON,#CXA+IDLE+CARL+CDR+CSTR+CSTP
        MOV     R6,#RXIDLE
        AJMP    I2CRTI
$EJ
;*************
; Handle STOP                           ;*RC*   Check for STOP first
;*************
RxStop: JNB     I2CON.STP,RxStart       ;Confirm it was Stop.
        MOV     I2CON,#CSTP             ;Clear it.
        JB      AA,RxStop0              ;Check end of message reached,
                                        ; Assert Acknowledge=1.
        ; Received STOP before end of message.
        SETB    MsgCheck                ;Signal interface error.
        SETB    I2CFG.MASTRQB
RxStop0: MOV    R6,#RXIDLE              ;Set context idle receiver.
        JNB     I2CON.ATN,RxStop1       ;If ATN, dispatch next event
        AJMP    DISPAT
RxStop1: MOV    I2CON,#IDLE             ;Become idle.
        AJMP    I2CRTI                  ;Resume interrupted activity

;**************
; Handle START
;**************
RxStart: JNB    I2CON.STR,RxFault       ;Was it start?
SlvStart: 
        MOV     I2CON,#CSTR+CARL        ;Yes, clear it.
        MOV     R4,#0                   ;ByteCnt=0
        AJMP    SetRXB                  ;Set up to receive byte.

; It wasn't DRDY, ARL, START, or STOP.  Inconsistency error.
RxFault: AJMP   PwrUp


$EJ
;***************************
; MASTER TRANSMITTER
;   R2=I2CDat, R3=BitCnt, R4=ByteCnt, R5=ATNCnt, R6=I2CCxt
;***************************
; Handle DRDY
;***************************
MASTER: MOV     A,R2                    ;Get data in A.
        JNB     I2CON.DRDY,MNDRDY       ;Is it DRDY?
; - context waiting to send bit
        CJNE    R6,#TXBIT,NTxBit        ;Context waiting to send bit?
Tx1:    MOV     I2DAT,A                 ;Send bit
Tx2:    RL      A                       ;Rotate the byte.
        DJNZ    R3,MN8Bit               ;Was it 8th bit?
        MOV     R6,#TXREAD              ;Set context waiting to read ACK
        JB      I2CON.DRDY,TxAck1       ;If DRDY, short cut.
        AJMP    WaitATN

        ; prepare next bit 1-7
MN8Bit: MOV     R2,A                    ;Put the data back.
        JB      I2CON.DRDY,Tx1          ;If DRDY, short cut.
        JB      I2CON.DRDY,Tx1
        JB      I2CON.DRDY,Tx1          ;One more try.
        AJMP    WaitATN 

; - context waiting to read ACK
NTxBit: CJNE    R6,#TXREAD,NTxAck1      ;Context waiting to read ACK?
TxAck1: MOV     I2CON,#CDR+CXA          ;Switch to receive mode.
        MOV     R6,#TXACK               ;Set context waiting for ACK.
        JB      I2CON.DRDY,NTxAck2      ;If DRDY, short cut.
        RL      A                       ;Align data in case ARL.
        DEC     R3
        MOV     R2,A
        AJMP    WaitATN

; - context waiting for ACK
NTxAck1: CJNE   R6,#TXACK,BeMast        ;Context waiting for ACK?
NTxAck2: MOV    A,I2CON                 ;Read from I2CON
        ANL     A,#80h                  ;Only need 7th bit
        JZ      AckOK
BadAck: ;Stop if negative ACK.
        DJNZ    NACnt,DoStp1            ;Keep pending msg till retry expires.
        AJMP    DoStp
        ;Ack Okay, prepare to send next byte.
AckOK:  INC     R4                      ;Point to byte we want to send.
        ;ACALL  Sample
        AJMP    DOTXB
        ;return should AJMP DNTXB
DNTXB:  MOV     A,R2                    ;Next byte in A.
        XRL     Check,A                 ;XOR with message check.
        MOV     R3,#8                   ;BitCnt=8
        MOV     R6,#TXBIT               ;Set context to send bit.
        JB      I2CON.DRDY,Tx1          ;If DRDY, short cut.
        AJMP    WaitATN

$EJ
;*******************************************************
; It wasn't DRDY.  Could be ARL, START, or STOP.
; If ARL, align data with tx bit that lost arbitration
; go to handle as SLAVE (MASTER was cleared).
; If it was START, we just became master.
; If STOP, not sure how this occured,
;   should have seen ARL, become idle SLAVE.
;*******************************************************
MNDRDY: JNB     I2CON.ARL,MNARL         ;ARL?
        RR      A                       ;Rotate data back.
        INC     R3                      ;Increment bit count.
        RR      A
        INC     R3
        MOV     R2,A
        AJMP    ARL0
MNARL:  JB      I2CON.STR,BeMast
        AJMP    IdleS

; - context START or not waiting for bit or ACK (from above).
; Must have just become master.  Start new message.
; If message is pending in send type, do it.
; Otherwise, check to send Position Report if needed.
BEMAST: MOV     R1,SndType              ;Get message type.
        CJNE    R1,#I_NoMsg,SndMsg      ;No message pending?
        JNB     MsgCheck,BeMast1        ;Send error?
        MOV     SndType,#I_Error
        CLR     MsgCheck
        AJMP    SndMsg
BeMast1: JB     SendRpt,PosMsg          ;Send position report?
        AJMP    DoStp                   ; then stop.

; Send Position report
PosMsg: JB      TxSelfRst,PosMsg1       ;If first user data,
        SETB    TxSelfRst
        MOV     SndType,#I_Reset        ;Send a I_Reset to my address.
        MOV     A,MyAddr
        MOV     NACnt,#1                ;No retry if not acknowledged.
        SJMP    SndMsg1
PosMsg1:MOV     SndType,#LD_Position    ;Send position report.
        CLR     SendRpt                 ;Clear pending report state.
        CLR     A                       ;Copy X-Y counts to xmt buffer.
        MOV     XBUF2,A                 ;Hi byte=0.
        MOV     YBUF2,A
        MOV     A,XCOUNT                ;Copy X.
        MOV     XBUF1,A
        JNB     ACC.7,Copy2             ;If negative,
        MOV     XBUF2,#0FFh             ; extend sign.
Copy2:  MOV     A,YCOUNT                ;Copy Y.
        MOV     YBUF1,A
        JNB     ACC.7,Copy3             ;If negative,
        MOV     YBUF2,#0FFh             ; extend sign.
Copy3:  CLR     A
        MOV     XCOUNT,A                ;Reset X-Y counters.
        MOV     YCOUNT,A
        CLR     Movement                ;Clear movement flag.
SndMsg: MOV     A,#Adr_Host
SndMsg1:MOV     I2DAT,A                 ;Send first bit by hand
        MOV     I2CON,#CARL+CSTR+CSTP   ;Clear start, release SCL
        ;Set context for rest of address
        MOV     Check,A                 ;Init checksum
        MOV     R2,A                    ;I2CDat=A
        MOV     R3,#8                   ;BitCnt=8
        MOV     R4,#0                   ;ByteCnt=0
        MOV     R6,#TXBIT               ;Set context waiting to send bit
        AJMP    Tx2
$EJ
; Completed sending message, do STOP
DoStp:  MOV     SndType,#I_NoMsg        ;Indicate cmd no longer pending.
DoStp1: ACALL   SndStop                 ;Send STOP.
        MOV     A,SndType               ;Is there a pending message?
        JNZ     DoStp3
        JB      SendRpt,DoStp3          ;Is there a Position Report?
        JNB     MsgCheck,DoStp4         ;Is there an error message?
DoStp3: ACALL   SDelay                  ;Yes, delay to give others
        ; a chance to become master without contention.
        SETB    I2CFG.MASTRQB           ;Request to be master again.
DoStp4: AJMP    RxStop0                 ;Borrow code from receiver.

; Send I2C STOP signal
SndStop:CLR     I2CFG.MASTRQB           ;Release Master request
        MOV     I2CON,#CDR+XSTP         ;Set to send stop
        JNB     I2CON.ATN,$             ;Wait for ATN
        MOV     I2CON,#CDR              ;Clear useless DRDY (rising SCL)
        JNB     I2CON.ATN,$             ;Wait for stop sent
        MOV     I2CON,#CARL+CSTP+CXA    ;Clear I2C bus
        MOV     R6,#RXIDLE              ;Set context idle receiver.
        RET

$EJ
;*******************************
; DO_RX_BYTE
; Received a complete byte, already acknowledged.
; Examine context to decide what to do with it.
;
; Enter: R2 (I2CDat) is byte received.
;        R4 is offset to byte just received.
; Exit:  Command parameters saved as needed.
;        Checksum verified.  Valid commands executed.
;        Return by AJMP DNRXB
;
;   R2=I2CDat, R3=BitCnt, R4=ByteCnt, R5=ATNCnt, R6=I2CCxt
;******************************
DORXB:  CJNE    R4,#0,DoRx1     ;Is it Address? (ByteCnt=0)?
        AJMP    DNRXB
DoRx1:  CJNE    R4,#1,DoRx2     ;Is it source Addr? (ByteCnt=1)
        AJMP    DNRXB
;
DoRx2:  MOV     A,R2            ;Get byte.
        CJNE    R4,#2,DoRx3     ;Is it P+len?
        RLC     A               ;Rotate Prot bit into C.
        MOV     Prot,C          ;Save it.
        MOV     A,R2            ;Get "len".
        ANL     A,#07Fh
        ADD     A,#3            ;Add overhead.
        MOV     MsgLen,A        ;Save message length.
        AJMP    DNRXB
;
DoRx3:  CJNE    R4,#3,DoRx4     ;Is it Command byte?
        MOV     RcvType,A       ;Save it
        AJMP    DNRXB
$EJ
;
; Test for end of command
;   If command has no data, byte offset 4 will be the checksum.
DoRx4:  MOV     A,MsgLen        ;Get message length.
        CJNE    A,ByteCnt,ToMny ;End of command?
                                ; sets carry if MsgLen<ByteCnt
        SETB    AA              ;Yes, do not Acknowledge more bytes.
        MOV     A,Check         ;Check in A
        JZ      CheckOk         ;Bad check?
        AJMP    RxErr
; Message check is Ok, dispatch valid commands
CheckOk: AJMP   DORXCMD         ;Return AJMP DNRXB
;
; Test for ByteCnt beyond message length
ToMny:  JNC     DoDat           ;Too many bytes?
        ; Yes, just exit, negative acknowledge already sent.
DoRx9:  AJMP    DNRXB

;
; Receive message data bytes
;   ByteCnt from 4 to (MsgLen-1)
;
; Branch on RcvType to decide what to do with each byte.
; Notice Reset, Identify, and Poll have no data.
;   Protocol: Assign New Address, Capabilities request
;
DoDat:  MOV     A,R2            ;Get data again since DoRx4 wiped it. 
        MOV     R7,RcvType      ;Put RcvType in R7 so we can CJNE
        JNB     Prot,DoRx9      ;Ignore device data stream.
        ;Process C/S command bytes.

$EJ
;***********************************************
; DORXB: Assign
;  Compare incoming bytes with ID string (ByteCnt=4-29).
;    If not equal, set not equal bit and ignore.
;  Compare ByteCnt 30-31 with random number.
;    If not equal, ignore (become idle receiver).
;  Save ByteCnt 32, the address as parameter (R0).
;***********************************************

Do5:    CJNE    R7,#I_AsgnAdr,Do6 ;Assign command?
        CJNE    R4,#4,Asn2      ;Start of ID string (ByteCnt=4)?
        MOV     R1,#0           ;Initialize R1 is index

        ; BytCnt <= 29
Asn4:   MOV     A,R1            ;Get offset
        ACALL   GET_ID
Asn5:   INC     R1              ;Increment for next
        CJNE    A,I2CDAT,AsnIg  ;Compare to received byte
        AJMP    DNRXB           ;Ok so far

Asn2:   CJNE    R4,#30,Asn3     ;ByteCnt=30?
        MOV     R1,#RandH       ;Yes, set to read random#
Asn3:   JC      ASN4            ;Jump if less than 30.

        ; ByteCnt >= 30
        CJNE    R4,#32,Asn6     ;ByteCnt=32?
        MOV     R0,I2CDat       ;Save new address in R0.
        AJMP    DNRXB
Asn6:   JNC     AsnIg           ;Jump if ByteCnt>32.
        MOV     A,@R1           ;Get byte of random #.
        SJMP    Asn5            ;Steal code from above.

AsnIg:  SETB    NotMyID         ;It's not for me
        AJMP    DNRXB

; Application Capabilities Request
Do6:    CJNE    R7,#I_CapReq,Do7        ;CapRequest?
        CJNE    R4,#4,Cpr1              ;ByteCnt=4?
        MOV     A,R2                    ;Check Cap hi pointer=0?
        JNZ     Cpr4                    ;No, reset length and offset to zero.
Cpr1:   CJNE    R4,#5,Cpr9              ;ByteCnt=5?
        ;Yes, check for valid offset.
        MOV     A,CapOffset             ;Compute next offset.
        ADD     A,CapLen
        CJNE    A,I2CDat,Cpr2           ;Send next?
        AJMP    Cpr8                    ; Yes, jump to set offset.
Cpr2:   MOV     A,R2                    ;A=received offset.
        JZ      Cpr8                    ;Send first?
        CJNE    A,CapOffset,Cpr4        ;Send previous?
        AJMP    DNRXB                   ; Yes, use current offset.
Cpr4:   CLR     A                       ;Reset Length and offset to zero.
        MOV     CapLen,A
Cpr8:   MOV     CapOffset,A             ;Load Cap offset.
Cpr9:   AJMP    DNRXB

; Set reporting Interval
Do7:    CJNE    R7,#LD_SetInterval,Do7a
        CJNE    R4,#4,Do7a              ;ByteCnt=4?
        MOV     R0,I2CDat               ;Save parameter in R0.
Do7a:   AJMP    DNRXB

; Command error (I_Error)
RxErr:  SETB    MsgCheck                ;Set to report error.
        SETB    I2CFG.MASTRQB
        AJMP    DNRXB
$EJ
;************************
; DO_RX_CMD
; Valid command received, do it.
; Dispatch commands based on RcvType
; Parameter value is in R0.
;
; Commands Recognized:  I_Reset, I_IdReq, I_AsgnAdr, I_CapReq,
;                       App_Test, App_Poll, App_SetInterval
;
; Return is:  AJMP DNRXB.  A is not preserved.
;***************************
DORXCMD:MOV     R7,RcvType              ;Put RcvType in R7 so we can CJNE.
        JB      Prot,DoC4               ;Go for Control/Status commands.
        AJMP    DNRXB                   ;Ignore device data stream msgs.

;Bus protocol commands
; Reset
DoC4:   CJNE    R7,#I_Reset,DoC5
        AJMP    PwrUp                   ;Do power up reset.
; Identify
DoC5:   CJNE    R7,#I_IdReq,DoC6
        MOV     SndType,#I_IdReply      ;Message type is identify
        JB      KeepID,RTBM             ;Keep same device number?
        SETB    KeepID
        MOV     RandL,TL                ;Random number <- T0
        MOV     RandH,TH
RTBM:   SETB    I2CFG.MASTRQB           ;Request to be master
        AJMP    DNRXB
; Assign
DoC6:   CJNE    R7,#I_AsgnAdr,DoC7
        JBC     NotMyID,NDo1            ;Was it a complete match?
        CJNE    R4,#33,RxErr            ;Check len=30+3
        CLR     TxSelfRst               ;Anticipate first user data.
        MOV     MyAddr,R0               ;Load new address
Ndo1:   AJMP    DNRXB
; Capabilities Request
DoC7:   CJNE    R7,#I_CapReq,DoC8
        CJNE    R4,#6,$+3               ;Check len>=3+3
        JC      RxErr
        MOV     SndType,#I_CapReply     ;Message type is Cap Report
        SETB    I2CFG.MASTRQB           ;Request to be master
        AJMP    DNRXB
; App Test
DoC8:   CJNE    R7,#App_Test,DoC9
        MOV     SndType,#App_TestReply  ;Send a test report
        SETB    I2CFG.MASTRQB           ;Request to be master.
        AJMP    DNRXB
; App Poll
DoC9:   CJNE    R7,#LD_Poll,DoC10
        MOV     A,MyAddr                ;Check for default address.
        CJNE    A,#Adr_Default,DoC9a
        SJMP    DoC9b                   ;Don't send Position reports
                                        ; to default address.
DoC9a:  SETB    SendRpt                 ;Flag to send Position report
        SETB    I2CFG.MASTRQB           ;Request to be master.
DoC9b:  AJMP    DNRXB

; Set Locator report interval
DoC10:  CJNE    R7,#LD_SetInterval,DoC11
        CJNE    R4,#5,$+3               ;Check len>=2+3
        JC      RxErr
        CJNE    R0,#0,DoC10a            ;Check parameter range.
        ; parameter=0, polling only. 
        MOV     TCON,#0                 ;Turn off Timer0.
        AJMP    DNRXB
DoC10a: CJNE    R0,#8,$+3
        JC      DoC10c                  ;Jump if R0<8.
        CJNE    R0,#26,$+3
        JNC     DoC10c                  ;Jump if R0>=26.
        ; 8 <= param <= 25, compute Timer0 reload value.
        MOV     A,#0FFh                 ;Start at FFFFh
        MOV     R1,A
DoC10b: CLR     C                       ;Loop to subtract
        SUBB    A,#MSECL                ; R0 milliseconds.
        XCH     A,R1
        SUBB    A,#MSECH
        XCH     A,R1
        DJNZ    R0,DoC10b
        MOV     RTL,A                   ;Set Timer0 reload.
        MOV     RTH,R1
        MOV     TCON,#INIT_TCON         ;Turn on Timer0.
        AJMP    DNRXB
DoC10c: AJMP    RxErr

; Unrecognized command (ignore it)
DoC11:  AJMP    DNRXB


$EJ
;******************
; DOTXB
; Transmitted a complete byte, has been acknowledged.
; Get next byte based on context.
; Enter:
;   R4 (ByteCnt) is the offset of the byte we wish to send.
; Exit:
;   R2 (I2CDat) is the next byte to transmit (if any).
;   A is not preserved.
;   Return via AJMP DNTXB.
;
;   R2=I2CDat, R3=BitCnt, R4=ByteCnt, R5=ATNCnt, R6=I2CCxt
;******************
; - Source address
DOTXB:  CJNE    R4,#1,DoTx1     ;ByteCnt=1?
        MOV     R2,MyAddr       ;Send source addr
        AJMP    DNTXB

; - Message length
DoTx1:  CJNE    R4,#2,DoTx2      ;ByteCnt=2?
        MOV     R2,#082h         ;Use 2 as default length,
                                 ; P=1, Control/Status msg.
        MOV     MsgLen,#5        ;Include overhead
        ; Compute length based on message type.
        ; If not Position Report, Identify, Output Error,
        ; or Reset, the length is 2.
        MOV     R7,SndType
; Position report
        CJNE    R7,#LD_Position,TxA ;Position report?
        MOV     R2,#6           ;Data length is 6 (P=0, data stream).
        MOV     MsgLen,#9       ;6 plus 3 overhead.
        AJMP    DNTXB
; Attention
TxA:    CJNE    R7,#I_Attn,TxI
        MOV     A,SelfTest      ;Check for ROM error
        CJNE    A,#ROM_ERROR,TxA9
Len3:   MOV     R2,#083h        ;Use Len=3 to include checksum.
        MOV     MsgLen,#6
TxA9:   AJMP    DNTXB
; Identify Reply
TxI:    CJNE    R7,#I_IdReply,TxC
        MOV     R2,#(80h+29)    ;Length for Identify.
        MOV     MsgLen,#(29+3)  ;Add overhead for MsgLen
        AJMP    DNTXB
; Capabilities Report
TxC:    CJNE    R7,#I_CapReply,TxE
        MOV     A,#CapFragLen           ;Get default fragment length.
        MOV     CapLen,A                ;Save it.
        ADD     A,CapOffset             ;Find end of fragment.
        CLR     C
        SUBB    A,#(CAP_END-CAP_START)  ;Is it beyond end of Cap String?
        JC      TxC3                    ;No, use default length.
TxC1:   CPL     A                       ;Yes, shorten as needed.
        INC     A
        ADD     A,Caplen
        MOV     CapLen,A
TxC3:   MOV     A,CapLen        ;Get fragment length.
        ADD     A,#083h         ;Compute data length.
        MOV     R2,A            ;Prepare to send it.
        ADD     A,#083h         ;Add 3 overhead for MsgLen (clear C/S).
        MOV     MsgLen,A
        AJMP    DNTXB
; Checksum or message framing error
TxE:    CJNE    R7,#I_Error,TxR
        SJMP    TxR1
; Reset
TxR:    CJNE    R7,#I_Reset,TxU
TxR1:   MOV     R2,#081h        ;Length for Reset (80+1).
        MOV     MsgLen,#4       ;1 plus 3 overhead.
TxU:    AJMP    DNTXB

; - Command code
DoTx2:  CJNE    R4,#3,DoTxLast          ;ByteCnt=3?
        MOV     R2,SndType              ;Send command code
        CJNE    R2,#LD_Position,TCC1    ; unless it is position report.
        MOV     R2,ReportBuf            ;In that case send 1st byte of report.
TCC1:   AJMP    DNTXB
;
; - Test for last byte of command (message check)
DoTxLast: MOV   A,MsgLen
        CJNE    A,ByteCnt,DoTxEnd ;Last byte of command?
                                ; sets Carry if A<ByteCnt 
        MOV     R2,Check        ;Yes, send check.
        AJMP    DNTXB
;
; - Test for beyond last byte of command
DoTxEnd: JNC    DoTx3           ;Beyond last byte (check)?
        MOV     NACnt,#5        ;Reset Negative Ack retry count.
        AJMP    DoStp           ;Send STOP.


; Transmit message data bytes
; ByteCnt from 3 to (length+2)
; Dispatch based on command type

DoTx3:  MOV     R7,SndType
        CJNE    R7,#LD_Position,DoT3    ;Position report?
        MOV     A,ByteCnt               ;Yes, send next byte.
        ADD     A,#ReportBuf-3
        MOV     R1,A
        MOV     I2CDat,@R1              ;R2=I2CDat=@R1
        AJMP    DNTXB

;Attention report
DoT3:   CJNE    R7,#I_Attn,DoT4
        MOV     R2,SelfTest             ;Send Power-up selftest and attention
        CJNE    R4,#5,AT4               ;ByteCnt=5?
        MOV     R2,RomSum               ;Yes, send checksum byte.
AT4:    AJMP    DNTXB
;
;Application test report
DoT4:   CJNE    R7,#App_TestReply,DoT5
        MOV     R2,SelfTest             ;Send Selftest result
        AJMP    DNTXB

;*******************************
; Identify report
;  Send ID string for ByteCnt 4-29 (26 bytes, last two are FFh).
;  Send random number for ByteCnt 30 and 31.
;*******************************
DoT5:   CJNE    R7,#I_IdReply,DoT6
        CJNE    R4,#4,IDR2      ;First byte (ByteCnt=4)?
        ;yes, set up to send ID string
        MOV     R1,#0           ;R1 is index

IDR4:   MOV     A,R1            ;Get offset
        ACALL   GET_ID
IDR5:   MOV     R2,A            ;Prepare to send it
        INC     R1              ;Increment for next
        AJMP    DNTXB
;
IDR2:   CJNE    R4,#30,IDR3     ;ByteCnt=30?
        MOV     R1,#RandH       ;Set to send random #.
IDR3:   JC      IDR4            ;Jump if less than 30.
        ; BytCnt >= 30
        CJNE    R4,#32,$+3
        JNC     IDR7            ;Jump if >= 32.
        MOV     A,@R1           ;Get byte of random #.
        SJMP    IDR5
IDR7:   AJMP    DoStp           ;Error! Beyond last byte,
                                ; STOP now.
;
;*******************************
; Capability report
;  Send next byte of capability string.
;  Uses R1 as index.
;*******************************
DoT6:   CJNE    R7,#I_CapReply,DoT7
        CJNE    R4,#4,Cap1      ;First byte (ByteCnt=4)?
        MOV     R2,#0           ;Send High byte of Offset (always 0).
        AJMP    DNTXB

Cap1:   CJNE    R4,#5,CAP2      ;Second byte (Offsetlo)?
        MOV     R2,CapOffset    ;Send Low byte of Offset.
        MOV     R1,CapOffset    ;Initialize R1 to use as index.
        AJMP    DNTXB

Cap2:   MOV     A,R1            ;Get Capabilities Character.
        ACALL   GET_CAP
        MOV     R2,A            ;Prepare to send it.
        INC     R1              ;Increment for next Character.
Cap3:   AJMP    DNTXB
;
; Unknown: How can we not know what we're sending?
DoT7:   AJMP    DNTXB


$EJ
;**********************
; GET_ID
; Get byte of ID string
;   Enter: offset of desired byte in A.
;   Exit:  A is the desired byte.
;**********************
GET_ID: INC     A               ;Skip RET
        MOVC    A,@A+PC         ;Get the byte
        RET
;ID string is defined here, length is 25
        DB      'A'             ;Protocol revision
        DB      'V1.1   '       ;Module revision
        DB      'DEC     '      ;Vendor name
        DB      'VSXXX-BB'      ;Module name
        DB      0FFh            ;1st byte of device #
        DB      0FFh            ;2nd byte of device #

;**********************
; GET_CAP
; Get byte of Capabilities string.
; This implementation supports up to 254 bytes only!
;   Enter: offset of desired byte in A.
;   Exit:  A is the desired byte.
;***********************
GET_CAP: INC    A               ;Skip RET
        MOVC    A,@A+PC         ;Get the byte
        RET
;Capabilities string is defined here.
CAP_START: DB   '('
        DB      ' prot(locator)'
        DB      ' type(mouse)'
        DB      ' buttons(1(L)2(R)3(M))'
        DB      ' dim(2)'
        DB      ' rel'
        DB      ' res(200 inch)'
        DB      ' range(-127 127)'
        DB      ' d0(dname(X))'
        DB      ' d1(dname(Y))'
        DB      ')'
CAP_END: DB     0       ;Null terminator (not used).
; Capabilities length is 121 bytes

        END
